/*********************************************************************************\
*     File:                  G S B L U R . C P P                                  *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * gsblur.cpp: gaussian blur filter
 *
 *  Contains:
 *         void GaussianBlur();        // Blur a image !
 *         void doGaussianConv24();    // Color implementation
 *         void doGaussianConvGrey();  // Greyscale implementation
 */

#include "ImgBasic.h"
#include "..\miguelan\miguelan.h"
#include <math.h>

/*********************************************************************************/
void ImgBasic::GaussianBlur(void)
{
  /* This funcition calls doGaussianConv24 if the image is coloured  *
   * or doGaussianConvGrey if the image is monocrome                 */

  byte *thepic;         /* Copy of the formatted original image */
  byte *tmpPic;         /* Buffer where store transformed image */
  char buf[64] = "3";   /* Buffer for user input                */
  int  i;               /* Counter                              */  
  float sigma;          /* Gaussian parameter                   */

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Filtro de suavizado gausiano:"
                  "\nIntroduce la caracterstica sigma",
                  buf, 64, "0123456789.", 1);
  
  /* Check for right values                                      */
  if (strlen(buf)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  sigma = (float) atof(buf);
  if (sigma<0.1) {
    Error("El valor de sigma es muy pequeno (no se produce suavizado)");
    return;
  }

  /* Call a function depending of the color information */

  if (mono) { // MONOCHROME IMAGE

    /* Get a standard greyscale version of the real image    */
    if (startGreyAlg(&thepic, &tmpPic)) 
	    return;      /* Failed to get the greyscale image    */

    /* Call the actual algorithm for greyscale images        */ 
    doGaussianConvGrey(thepic, tmpPic, pWIDE, pHIGH, sigma);

    /* Store new image, allow 'undo', uptade display, etc.   */
    endGreyAlg(thepic, tmpPic);
    
  } else {  // COLOR IMAGE

    /* Get a standard 24-bit RGB version of the real image   */
    if (start24bitAlg(&thepic, &tmpPic)) 
	    return; /* Failed to get the RGB-24bit image         */
    
    /* Call the actual algorithm for color images            */ 
    doGaussianConv24(thepic, tmpPic, pWIDE, pHIGH, sigma);

    /* Store new image, allow 'undo', uptade display, etc.   */
    end24bitAlg(thepic, tmpPic);    
  }

}   /* End of GaussianBlur image                             */


/*********************************************************************************/
void ImgBasic::doGaussianConv24   ( byte *pic24, byte *results,
                                     int w, int h,float sigma )
{
  register int i;
  register byte *src, *dst;
  byte *aux_src, *aux_dst;
  
  /* ALLOCATE MEMORY */
  aux_src = (byte*) malloc( (long)w*(long)h );
  if (!aux_src) {
    Error("Fallo al requerir memoria");
    return;
  }
  aux_dst = (byte*) malloc( (long)w*(long)h );
  if (!aux_dst) {
    free(aux_src);
    Error("Fallo al requerir memoria");
    return;
  }

  /* BLUR RED COMPONENT! */
  src=pic24; dst=aux_src;
  for(i=w*h ; i; i--, src+=3) *dst++ = *src;
  doGaussianConvGrey(aux_src, aux_dst, w, h, sigma);
  src=aux_dst; dst=results;
  for(i=w*h ; i; i--, dst+=3, src++) *dst = *src;
  
  /* BLUR GREEN COMPONENT! */
  src=pic24+1; dst=aux_src;
  for(i=w*h ; i; i--, src+=3) *dst++ = *src;
  doGaussianConvGrey(aux_src, aux_dst, w, h, sigma);
  src=aux_dst; dst=results+1;
  for(i=w*h ; i; i--, dst+=3, src++) *dst = *src;

  /* BLUR BLUE COMPONENT! */
  src=pic24+2; dst=aux_src;
  for(i=w*h ; i; i--, src+=3) *dst++ = *src;
  doGaussianConvGrey(aux_src, aux_dst, w, h, sigma);
  src=aux_dst; dst=results+2;
  for(i=w*h ; i; i--, dst+=3, src++) *dst = *src;

  /* FREE MEMORY */
  free(aux_src);
  free(aux_dst);
}

/*********************************************************************************/
void ImgBasic::doGaussianConvGrey   ( byte *picGrey, byte *gausiana,
                                        int columnas,int filas, float sigma )
{
  /* Blur grey imagen with gaussian weight: f=exp(-xx/2ss) */  
  float far *aux;      /* Auxiliar image for separating components */
  int indice1,indice2,ancho,ancho_max;
  register int posicion1,posicion2;
  float valor_1,factor;
  float gaus[MAX_SIGMA];

  /* Allocate memory */
  aux = (float*) malloc( (long)filas*(long)columnas*sizeof(float) );
  if (!aux) {
    Error("Fallo al requerir memoria");
    return;
  }

  /* width of gaussian window */
  ancho_max=(int)(2.0*sigma+1);
  ancho_max=MIN(ancho_max,MAX_SIGMA);
  
  /* gaussian weights (only half window because of simetry) */
  gaus[0]=(float)1;
  for ( indice1=1 ; indice1<ancho_max ; indice1++ )
     gaus[indice1]=(float)exp(-indice1*indice1/(2*sigma*sigma));

  /* Scale factor */
  factor=gaus[0];
  for ( ancho=1 ; ancho<ancho_max ; ancho++)
       factor += 2*gaus[ancho];
  factor *= factor; //factor^2 => saves fil*col multipic

  /* BLUR X (calculated with a "mirror image" beyond the border )*/
  for ( indice1=0 ; indice1<filas ; indice1++ ) {
     for ( indice2=0 ; indice2<columnas ; indice2++ ) {
       valor_1=picGrey[indice2+indice1*columnas];
       for ( ancho=1 ; ancho<ancho_max ; ancho++) {	
         /* Don't excess limits! */
         posicion1=indice2+ancho;
         posicion1= (posicion1<0)         ? (-posicion1) :
                    (posicion1>=columnas) ? (posicion1-columnas) :
                                            posicion1;
		     posicion2=indice2-ancho;
         posicion2= (posicion2<0)         ? (-posicion2) :
                    (posicion2>=columnas) ? (posicion2-columnas) :
                                            posicion2;
         valor_1 += ( picGrey[posicion1+indice1*columnas] +
			          picGrey[posicion2+indice1*columnas] ) * gaus[ancho];
       }
       aux[indice2 + indice1*columnas]=valor_1;
     }
  }
  /* BLUR Y (calculated with a "mirror image" beyond the border )*/
  for ( indice1=0 ; indice1<columnas ; indice1++ ) 
     for ( indice2=0 ; indice2<filas ; indice2++ ) {
       valor_1=aux[indice1+indice2*columnas];
       for ( ancho=1 ; ancho<ancho_max ; ancho++) {
         /* Don't excess limits! */
         posicion1 =indice2+ancho;
         posicion1 = (posicion1<0)         ? (-posicion1) :
                     (posicion1>=filas)    ? (posicion1-filas) :
                                              posicion1;
         posicion2 =indice2-ancho;
         posicion2 = (posicion2<0)         ? (-posicion2) :
                     (posicion2>=filas)    ? (posicion2-filas) :
                                              posicion2;
         valor_1 += (aux[indice1+posicion1*columnas] +
					 aux[indice1+posicion2*columnas] ) * gaus[ancho];
       }
       posicion1 = (int) (valor_1/factor);
       gausiana [indice1 + indice2*columnas]=(byte) (posicion1 > 255) ? 255 :
                                                    (posicion1 < 0  ) ? 0   :
                                                     posicion1;
     }
  
  /* FREE MEMORY */
  free(aux);
}
/*********************************************************************************/
