/*********************************************************************************\
*     File:                  S O B E L . C P P                                    *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * sobel.cpp: algoritms to calcule gradient ( Sobel window )
 *
 *  Contains:
 *         void SobelGradient();
 *         void saturePic24();
 *         void saturePicGrey();
 */

#include "ImgBasic.h"
#include <time.h>

/*********************************************************************************/
void ImgBasic::saturePicGrey(byte* tmpPic, int pWIDE, int sx, int sy,
                              int sw, int sh, float factor)
{
   register byte *ptr;
   register byte maxv=0;
   int i,j;

   /* Normalize results */
   for (i=sy, maxv=0; i<sy+sh; i++) {              /* compute max value */
     ptr = tmpPic + i*pWIDE + sx;
     for (j=sw; j; j--, ptr++) {
       if (*ptr>maxv) maxv = *ptr;
     }
   }

   if (maxv>0) {
     short scale = (short) (((float) maxv)/factor);
     for (i=sy; i<sy+sh; i++) {            /* normalize */
       ptr = tmpPic + i*pWIDE + sx;
       for (j=sw; j; j--) {
         register short v =
             (((short) *ptr) << 8 ) / scale;   /*  aprox * 255 = <<8 */
         *ptr++ =  (v & 0xFF00) ? (byte)255 : (byte) v;
       }
     }
   }

}

/*********************************************************************************/
void ImgBasic::saturePic24(byte* tmpPic, int pWIDE, int sx, int sy,
                            int sw, int sh, float factor)
{
   register byte *ptr;
   register byte maxv=0;
   int i,j;

   /* Normalize results */
   for (i=sy, maxv=0; i<sy+sh; i++) {              /* compute max value */
     ptr = tmpPic + (i*pWIDE + sx) * 3;
     for (j=sx; j<sx+sw; j++, ptr+=3) {
       register byte v = MONO(ptr[0], ptr[1], ptr[2]);
       if (v>maxv) maxv = v;
     }
   }

   if (maxv>0) {
     short scale = (short) (((float) maxv)/factor);
     for (i=sy; i<sy+sh; i++) {            /* normalize */
       ptr = tmpPic + (i*pWIDE + sx) * 3;
       for (j=0; j<sw*3; j++) {
         register short v =
             (((short) *ptr) << 8 ) / scale;   /*  aprox * 255 = <<8 */
         *ptr++ =  (v & 0xFF00) ? (byte)255 : (byte) v;
       }
     }
   }
}


/*********************************************************************************/
void ImgBasic::SobelGradient(void)
{
  byte *thepic, *tmpPic;
  char *str;
  int sx,sy,sw,sh;

  WaitCursor();

  if (HaveSelection()) GetSelRCoords(&sx,&sy,&sw,&sh);
  else { sx = 0;  sy = 0;  sw = pWIDE;  sh = pHIGH; }
  CropRect2Rect(&sx,&sy,&sw,&sh,0,0,
                pWIDE,pHIGH);

  str = "Doing edge detection...";
  SetISTR(ISTR_INFO, str);

  if (mono) { // GRAY IMAGE
      if (startGreyAlg(&thepic, &tmpPic)) return;
      
      xvbcopy((char *) thepic, (char *) tmpPic,
              (size_t) (pWIDE*pHIGH));

      doEdgeConvolvGrey(thepic, pWIDE, pHIGH, tmpPic, sx,sy,sw,sh);

      saturePicGrey(tmpPic, pWIDE,sx,sy,sw,sh);
  
      SetISTR(ISTR_INFO, "%snormalizing...", str);

      endGreyAlg(thepic, tmpPic);
    
  } else {  // COLOR IMAGE
      
      if (start24bitAlg(&thepic, &tmpPic)) return;
      xvbcopy((char *) thepic, (char *) tmpPic,
        (size_t) (pWIDE*pHIGH*3));

      doEdgeConvolv24(thepic, pWIDE, pHIGH,
                      tmpPic, sx,sy,sw,sh);
  
      SetISTR(ISTR_INFO, "%snormalizing...", str);

      saturePic24(tmpPic,pWIDE,sx,sy,sw,sh);

      end24bitAlg(thepic, tmpPic);
  }
  
  ArrowCursor();
}


/*********************************************************************************/
void ImgBasic::doEdgeConvolv24( byte *pic24, int w, int h,  byte *results, int selx,
                          int sely, int selw, int selh)
{

  /* convolves with two edge detection masks (vertical and horizontal)
     simultaneously, taking Max(abs(results)) 
     
     The two masks are (hard coded):

          -1 0 1             -1 -2 -1
      H = -2 0 2     and V =  0  0  0
          -1 0 1              1  2  1

     divided into 
           -1 0 0         0 0 0         0 0 1        0  1 0
       a =  0 0 0 ,  b = -1 0 1 ,  c =  0 0 0 ,  d = 0  0 0 .
            0 0 1         0 0 0        -1 0 0        0 -1 0

     So H = a + 2*b + c,  V = a - c - 2*d.
     gradient = max(abs(H),abs(V)).
          
     Also, only does pixels in which the masks fit fully onto the picture
     (no pesky boundary conditionals)  */


  register byte *p24;
  register int   bperlin, a, b, c, d, rsum, gsum, bsum;
  byte          *rp;
  int            x, y;


  
  bperlin = w * 3;

  for (y=sely+1; y<sely+selh-1; y++) {
    ProgressMeter(sely+1, (sely+selh-1)-1, y, "Edge Detect");
    if ((y & 63) == 0) WaitCursor();

    rp  = results + (y*w + selx+1)*3;
    p24 = pic24   + (y*w + selx+1)*3;

    for (x=selx+1; x<selx+selw-1; x++, p24+=3) {
      /* RED PLANE */
      a = p24[bperlin+3]  - p24[-bperlin-3]; /* bottom right - top left */
      b = p24[3]   -        p24[-3];         /* mid    right - mid left */
      c = p24[-bperlin+3] - p24[bperlin-3];  /* top    right - bottom left */
      d = p24[-bperlin]   - p24[bperlin];    /* top    mid   - bottom mid */

      rsum = a + (b<<1) + c;      /* horizontal gradient H = a + 2*b + c*/
      if (rsum < 0) rsum = -rsum;
      a = a - c - (d<<1);         /* vertical gradient V = a - c - 2*d*/
      if (a < 0) a = -a;
      if (a > rsum) rsum = a;
      //rsum >>= 2;

      /* GREEN PLANE */
      a = p24[bperlin+4]  - p24[-bperlin-2]; /* bottom right - top left */
      b = p24[4]   -        p24[-2];         /* mid    right - mid left */
      c = p24[-bperlin+4] - p24[bperlin-2];  /* top    right - bottom left */
      d = p24[-bperlin+1] - p24[bperlin+1];  /* top    mid   - bottom mid */

      gsum = a + (b<<1) + c;      /* horizontal gradient */
      if (gsum < 0) gsum = -gsum;
      a = a - c - (d<<1);         /* vertical gradient */
      if (a < 0) a = -a;
      if (a > gsum) gsum = a;
      //gsum >>= 2;

      /* BLUE PLANE */
      a = p24[bperlin+5]  - p24[-bperlin-1]; /* bottom right - top left */
      b = p24[5]   -        p24[-1];         /* mid    right - mid left */
      c = p24[-bperlin+5] - p24[bperlin-1];  /* top    right - bottom left */
      d = p24[-bperlin+2] - p24[bperlin+2];  /* top    mid   - bottom mid */

      bsum = a + (b<<1) + c;      /* horizontal gradient */
      if (bsum < 0) bsum = -bsum;
      a = a - c - (d<<1);         /* vertical gradient */
      if (a < 0) a = -a;
      if (a > bsum) bsum = a;
      //bsum >>= 2;

      //*rp++ = (byte)  rsum; *rp++ = (byte) gsum; *rp++ = (byte) bsum;
      
      *rp++ = (byte) ((rsum>255)? 255 : rsum);
      *rp++ = (byte) ((gsum>255)? 255 : gsum);
      *rp++ = (byte) ((bsum>255)? 255 : bsum);
    }
  }
}

/*********************************************************************************/
void ImgBasic::doEdgeConvolvGrey( byte *pic8, int w, int h,  byte *results, int selx,
                          int sely, int selw, int selh)
{
  /* convolves with two edge detection masks (vertical and horizontal)
     simultaneously, taking Max(abs(results)) 
     
     The two masks are (hard coded):

          -1 0 1             -1 -2 -1
      H = -2 0 2     and V =  0  0  0
          -1 0 1              1  2  1

     divided into 
           -1 0 0         0 0 0         0 0 1        0  1 0
       a =  0 0 0 ,  b = -1 0 1 ,  c =  0 0 0 ,  d = 0  0 0 .
            0 0 1         0 0 0        -1 0 0        0 -1 0

     So H = a + 2*b + c,  V = a - c - 2*d.
     gradient = max(abs(H),abs(V)).
          
     Also, only does pixels in which the masks fit fully onto the picture
     (no pesky boundary conditionals)  */

  register byte *pGrey;
  register int   a, b, c, d, sum;
  byte          *rp;
  int            x, y;


  for (y=sely+1; y<sely+selh-1; y++) {
    ProgressMeter(sely+1, (sely+selh-1)-1, y, "Edge Detect");
//    if ((y & 63) == 0) WaitCursor();
    rp    = results + (y*w + selx+1);
    pGrey = pic8 + (y*w + selx+1);

    for (x=selx+1; x<selx+selw-1; x++, pGrey++) {
      
      a = pGrey[w+1]  - pGrey[-w-1]; /* bottom right - top left */
      b = pGrey[1]   -  pGrey[-1];   /* mid    right - mid left */
      c = pGrey[-w+1] - pGrey[w-1];  /* top    right - bottom left */
      d = pGrey[-w]   - pGrey[w];    /* top    mid   - bottom mid */

      sum = a + (b<<1) + c;      /* horizontal gradient H = a + 2*b + c*/
      if (sum < 0) sum = -sum;
      a = a - c - (d<<1);         /* vertical gradient V = a - c - 2*d*/
      if (a < 0) a = -a;
      if (a > sum) sum = a;
      //rsum >>= 2;

      *rp++ = (byte) ((sum>255)? 255 : sum );
    }
  }
}
