/*********************************************************************************\

*     File:                  C O N 2 4 T O 8 . C P P                              *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * con24to8.cpp  -  contains the 24-to-8-bit Conv24to8() procedure
 *                  and the 8-to-24-bit Conv8to24() procedure
 *
 * The Conv24to8 procedure takes a pointer to a 24-bit image (loaded
 * previously).  The image will be a w * h * 3 byte array of
 * bytes.  The image will be arranged with 3 bytes per pixel (in order
 * R, G, and B), pixel 0 at the top left corner.  (As normal.)
 * The procedure also takes a maximum number of colors to use (numcols)
 * and pointers to three 256-long arrays of bytes (to hold the returned
 * colormap)
 *
 * Note that Conv24to8() does NOT free the pic24 image under any circumstances
 *
 * The Conv24to8 procedure will set up the following:  it will allocate, make
 * & return 'pic8', a 'w' by 'h' (passed in values) 8-bit picture.
 * it will load up the rmap, gmap and bmap colormap arrays.  it will NOT 
 * calculate numcols, since the cmap sort procedure has to be called anyway
 *
 * Conv24to8 returns 'pic8' if successful, 'NULL' on failure (presumably on a 
 * malloc())
 *
 * The 'slow' code, while still based on Heckbert's Median Cut algorithm, 
 * has been shamelessly lifted from the Independent JPEG Group's software
 * (jquant2.c), as (for a variety of reasons) theirs was far better than 
 * the version I was previously using.  Thanks guys!
 *
 * Also, as is my way, I've stripped out most of the IJG's well-written
 * comments regarding their algorithm.  Folks interested in learning how it
 * works are encouraged to look at the original source.  (jpeg/jquant2.c)
 *
 */

/*
 * Portions Copyright (C) 1989, 1991 by Jef Poskanzer.  See copyright notice
 * below, at the beginning of the relevant code.
 */

#include "IOfile.h"


typedef struct box* box_vector;

struct box {
  int index;
  int colors;
  int sum;
};



static int    quick_check (const byte*, int,int, byte*, byte*,byte*,byte*,int);
static int    quick_quant (const byte*, int,int, byte*, byte*,byte*,byte*,int);

int    quick_check_map(const byte *pic24, int w,int h, byte *rmap,
                       byte *gmap,byte *bmap, int maxcol);
int    GetSlowMapFromPIC24(const byte *pic24,int w,int h,int nc,
                           byte *rm,byte *gm,byte *bm);

/* *************** stuff  in QUANT.CPP ************** */

extern int    slow_quant  (const byte *pic24, int w, int h, byte *pic8, byte *rm,
                           byte *gm, byte *bm, int descols);

extern int    slow_quant_map  (const byte *pic24, int w, int h, byte *rm,
                           byte *gm, byte *bm, int descols);

/****************************/
byte *IOfile::Conv24to8(const byte *pic24,int w,int h,
                           byte *rm,byte *gm,byte *bm,int nc)
{
  /* returns pointer to new 8-bit-per-pixel image (w*h) if successful, or
     NULL if unsuccessful */
  
  int   i;
  byte *pic8;
  
  if (!pic24) return NULL;

  pic8 = (byte *) malloc((size_t) (w * h));
  if (!pic8) {
    Error("Conv24to8() - fallo pidiendo memoria (malloc) 'pic8'\n");
    return pic8;
  }

  if (nc<=0) nc = 256;  /* 'nc == 0' breaks code */

  if (quick_check(pic24, w,h, pic8, rm,gm,bm, nc)) { 
    SetISTR(ISTR_INFO,"No color compression was necessary.\n");
    return pic8;   
  }

  switch (conv24) {
  case CONV24_FAST:
    SetISTR(ISTR_INFO,"Doing 'quick' 24-bit to 8-bit conversion.");
    i = quick_quant(pic24, w, h, pic8, rm, gm, bm, nc);
    break;
    
  case CONV24_BEST:
    SetISTR(ISTR_INFO,"Doing 'best' 24-bit to 8-bit conversion.");
    i = ppm_quant(pic24, w, h, pic8, rm, gm, bm, nc);
    break;
    
  case CONV24_SLOW:
  default:
    SetISTR(ISTR_INFO,"Doing 'slow' 24-bit to 8-bit conversion.");
    i = slow_quant(pic24, w, h, pic8, rm, gm, bm, nc);
    break;
  }
  
  if (i) { 
   free(pic8);
   pic8 = NULL;
   Error("Fallo en el algoritmo Conv24to8");
 }
  return pic8;
}

/***************************************************************/
byte *IOfile::Conv8to24(const byte *pic8, int w, int h, 
                           const byte *rmap,const byte *gmap,const byte *bmap)
{
  /* converts an w*h 8-bit image (with colormap rmap,gmap,bmap) into a
   * 24-bit image.  Note, 'pic8' could be NULL
   *
   * returns pointer to new 24-bits-per-pixel image (w*h) if successful,
   * or NULL if unsuccessful
   */

  register int   i;
  byte *pic24;
  register byte *sp, *dp;
  pic24 = (byte *) malloc((size_t) (w * h * 3));
  if (!pic24) {
    FatalError("Conv8to24: fallo requiriendo memoria (malloc)");
    return pic24;
  }
  
  WaitCursor();
  
  for (i=w*h, sp=(byte*)pic8, dp=pic24; i; i--, sp++) {
    //if ((i&0x1ffff)==0) WaitCursor();
    *dp = rmap[*sp];
    dp++;
    *dp = gmap[*sp];
    dp++;
    *dp = bmap[*sp];
    dp++;
  }

  ArrowCursor();
  return pic24;
}

/***************************************/

byte *IOfile::ConvGreyto24(const byte *pic8, int w, int h)
{
  /* converts an w*h greyscale image into a
   * 24-bit image.  Note, 'pic8' could be NULL
   *
   * returns pointer to new 24-bits-per-pixel image (w*h) if successful,
   * or NULL if unsuccessful
   */

  register int   i;
  byte *pic24;
  register byte *sp, *dp;

  pic24 = (byte *) malloc((size_t) (w * h * 3));
  if (!pic24) {
    FatalError("ConvGrayto24: fallo pidiendo memoria (malloc)");
    return pic24;
  }

  WaitCursor();
  
/*
  for (i=w*h, sp=(byte*)pic8, dp=pic24-1; i; i--, sp++) {
    *(++dp) = *(++dp) = *(++dp) = *sp;
  }
*/

  for (i=w*h, sp=(byte*)pic8, dp=pic24; i; i--, sp++) {
    *dp=*sp;
    dp++;
    *dp=*sp;
    dp++;
    *dp=*sp;
    dp++;    
  }

  ArrowCursor();

  return pic24;
}

/***************************************/
byte *IOfile::Conv24toGrey(const byte *pic24,int w,int h)
{
  /* returns pointer to new 8-bit-per-pixel grayscale image (w*h) if 
     successful, or NULL if unsuccessful */
  
  int   i;
  byte *pic8, *p, *ep;
  
  if (!pic24) return NULL;

  pic8 = (byte *) malloc((size_t) (w * h));
  if (!pic8) {
    Error("Conv24to8() - fallo pidiendo memoria (malloc) 'pic8'\n");
    return pic8;
  }

  for (i=w * h, p=pic8,ep=pic; i; i--, p++, ep+=3)
          *p = MONO(ep[0],ep[1],ep[2]);
  
  return pic8;
}

/********************************/
byte *IOfile::Conv8toGrey(const byte *pic8, int w, int h, 
                           const byte *rmap,const byte *gmap,const byte *bmap)
{
  /* converts an w*h 8-bit image (with colormap rmap,gmap,bmap) into a
   * 24-bit image.  Note, 'pic8' could be NULL
   *
   * returns pointer to new 24-bits-per-pixel image (w*h) if successful,
   * or NULL if unsuccessful
   */

  int   i;
  byte *greypic, *p, *ep, grey[256];

  /* create a Greyscale direct mapped image*/
  greypic = (byte *) malloc((size_t) w * h);
  if (!greypic) {
    FatalError("Conv8toGrey: Ha fallado el alojamiento de la"
               "imagen en escala de grises");
    return NULL;
  }
  else { 
    for (i=0; i<256; i++)
      grey [i] = MONO(rmap[i], gmap[i], bmap[i]);
    /* the random colormap has to be converted into a linear, shorted grayscale */
    for (i=w*h, p=greypic, ep=(byte*) pic8; i; i--, p++, ep++) 
      *p = greypic[grey[*ep]];          
  }
  
  return greypic;
}

  

/*********************************************************************************/
void IOfile::RedComponent  (byte* component, const byte *pic24,
                            u_int wide, u_int high)
{
  if (!component) {
    //Don't do nothing if a null pointer is given
    Error("Puntero enviado a GreenComponent vaco.\n"
          "Reserva pWIDE*pHIGH memoria antes de llamar a la funcin");
    return;
  }
  register byte *src, *dst=component;
  register u_int i;
  if (pic24 != NULL) {
    /* Use pic24, wide & high */
    i=wide*high;
    src = (byte*) pic24;
  } else {
    /* Use IOfile's picture */
    i=pWIDE*pHIGH;
    src = pic;  
  }
  if (picType==PIC24 || pic24 ) 
      for( ; i; i--, src+=3) *dst++ = *src;
  else //PIC8
      for(; i; i--, src++) *dst++ = rMap[*src];
}

/*********************************************************************************/
void IOfile::GreenComponent (byte* component, const byte *pic24,
                            u_int wide, u_int high)
{
  if (!component) {
    //Don't do nothing if a null pointer is given
    Error("Puntero enviado a GreenComponent vaco.\n"
          "Reserva pWIDE*pHIGH memoria antes de llamar a la funcin");
    return;
  }
  register byte *src, *dst=component;
  register u_int i;
  if (pic24 != NULL) {
    /* Use pic24, wide & high */
    i=wide*high;
    src = (byte*) pic24;
  } else {
    /* Use IOfile's picture */
    i=pWIDE*pHIGH;
    src = pic;  
  }
  if (picType==PIC24 || pic24) {
      src++;
      for( ; i; i--, src+=3) *dst++ = *src;
  } else { //PIC8
      for(; i; i--, src++) *dst++ = gMap[*src];
  }
}
/*********************************************************************************/
void IOfile::BlueComponent (byte* component, const byte *pic24,
                            u_int wide, u_int high)
{
  if (!component) {
    //Don't do nothing if a null pointer is given
    Error("Puntero enviado a BlueComponent vaco.\n"
          "Reserva pWIDE*pHIGH memoria antes de llamar a la funcin");
    return;
  }
  register byte *src, *dst=component;
  register u_int i;
  if (pic24 != NULL) {
    /* Use pic24, wide & high */
    i=wide*high;
    src = (byte*) pic24;
  } else {
    /* Use IOfile's picture */
    i=pWIDE*pHIGH;
    src = pic;  
  }
  if (picType==PIC24 || pic24) {
      src += 2;
      for( ; i; i--, src+=3) *dst++ = *src;
  } else { //PIC8
      for(; i; i--, src++) *dst++ = bMap[*src];
  }
}
/*********************************************************************************/

int IOfile::RGBtoPIC24 (byte* Red, byte* Green, byte* Blue,
                        u_int wide, u_int high)
{
  byte *pic24;

  if (!(Red && Green && Blue)) {  //Check nul pointer 
    Error("Alguna componente RGB enviada a RGBtoPIC24 est vaca.");
    return -1;
  } 

  if (!wide && !high) { 
    // Default dimension in case wide = 0 && high = 0
    wide = pWIDE;
    high = pHIGH;
  } else if (!wide || !high) {
    Error("Dimensiones de la imagen es incorrecta en la funcin RGBtoPIC24");
    return -1;
  }

  pic24 = (byte *) malloc (wide*high*3);
  if (!pic24) {
    Error("Fallo al requerir memoria en RGBtoPIC24");
    return -1;
  }
  
  { // Copy RGB components to pic24
    register u_int i=wide*high;
    register byte *dst=pic24;
    for(; i; i--) {
      *dst++ = *Red++;
      *dst++ = *Green++;
      *dst++ = *Blue++;
    }
  }

  // Update image
  StoreNewPic( pic24, PIC24,FALSE);
  
  // Refresh display
  RedrawPic();

  //sucessfull return
  return 0;
}


/****************************/
static int quick_check(const byte *pic24, int w,int h, byte *pic8, byte *rmap,
                       byte *gmap,byte *bmap, int maxcol)
{
  /* scans picture until it finds more than 'maxcol' different colors.  If it
     finds more than 'maxcol' colors, it returns '0'.  If it DOESN'T, it does
     the 24-to-8 conversion by simply sticking the colors it found into
     a colormap, and changing instances of a color in pic24 into colormap
     indicies (in pic8) */

  unsigned long colors[256],col;
  int           i, nc, low, high, mid;
  byte         *p, *pix;

  if (maxcol>256) maxcol = 256;

  /* put the first color in the table by hand */
  nc = 0;  mid = 0;  

  for (i=w*h,p=(byte*)pic24; i; i--) {
    col  = (((u_long) *p++) << 16);  
    col += (((u_long) *p++) << 8);
    col +=  *p++;

    /* binary search the 'colors' array to see if it's in there */
    low = 0;  high = nc-1;
    while (low <= high) {
      mid = (low+high)/2;
      if      (col < colors[mid]) high = mid - 1;
      else if (col > colors[mid]) low  = mid + 1;
      else break;
    }

    if (high < low) { /* didn't find color in list, add it. */
      if (nc>=maxcol) return 0;
      xvbcopy((char *) &colors[low], (char *) &colors[low+1],
        (nc - low) * sizeof(u_long));
      colors[low] = col;
      nc++;
    }
  }


  /* run through the data a second time, this time mapping pixel values in
     pic24 into colormap offsets into 'colors' */

  for (i=w*h,p=(byte*)pic24, pix=pic8; i; i--,pix++) {
    col  = (((u_long) *p++) << 16);  
    col += (((u_long) *p++) << 8);
    col +=  *p++;

    /* binary search the 'colors' array.  It *IS* in there */
    low = 0;  high = nc-1;
    while (low <= high) {
      mid = (low+high)/2;
      if      (col < colors[mid]) high = mid - 1;
      else if (col > colors[mid]) low  = mid + 1;
      else break;
    }

    if (high < low) {
      debug_printf("quick_check:  impossible situation!\n");
      exit(1);
    }
    *pix = mid;
  }

  /* and load up the 'desired colormap' */
  for (i=0; i<nc; i++) {
    rmap[i] = (byte) (colors[i]>>16);  
    gmap[i] = (byte) (colors[i]>>8) & 0xff;
    bmap[i] = (byte) colors[i]      & 0xff;
  }

  return 1;
}




/************************************/
static int quick_quant(const byte *p24,int w,int h, byte *p8, byte *rmap,
                       byte *gmap,byte *bmap, int nc)
{
  /* called after 'pic8' has been alloced, pWIDE,pHIGH set up, mono/1-bit
     checked already */
  
/* up to 256 colors:     3 bits R, 3 bits G, 2 bits B  (RRRGGGBB) */
#define RMASK      0xe0
#define RSHIFT        0
#define GMASK      0xe0
#define GSHIFT        3
#define BMASK      0xc0
#define BSHIFT        6

  byte *pp;
  int  r1, g1, b1;
  int  *thisline, *nextline, *thisptr, *nextptr, *tmpptr;
  int  i, j, val, pwide3;
  int  imax, jmax;

  pp = p8;  pwide3 = w * 3;  imax = h-1;  jmax = w-1;


  /* load up colormap:
   *   note that 0 and 255 of each color are always in the map;
   *   intermediate values are evenly spaced.
   */

  for (i=0; i<256; i++) {
    rmap[i] = (((i<<RSHIFT) & RMASK) * 255 + RMASK/2) / RMASK;
    gmap[i] = (((i<<GSHIFT) & GMASK) * 255 + GMASK/2) / GMASK;
    bmap[i] = (((i<<BSHIFT) & BMASK) * 255 + BMASK/2) / BMASK;
  }
  

  thisline = (int *) malloc(pwide3 * sizeof(int));
  nextline = (int *) malloc(pwide3 * sizeof(int));
  if (!thisline || !nextline) {
    if (thisline) free(thisline);
    if (nextline) free(nextline);
    debug_printf("%s: unable to allocate memory in quick_quant()\n", cmd);
    return(1);
  }
  
  /* get first line of picture */
  for (j=pwide3, tmpptr=nextline; j; j--) *tmpptr++ = (int) *p24++;
  
  for (i=0; i<h; i++) {
    tmpptr = thisline;  thisline = nextline;  nextline = tmpptr;   /* swap */
    
    //if ((i&0x3f) == 0) WaitCursor();

    if (i!=imax)   /* get next line */
      for (j=pwide3, tmpptr=nextline; j; j--)
  *tmpptr++ = (int) *p24++;
    
    for (j=0, thisptr=thisline, nextptr=nextline; j<w; j++,pp++) {
      r1 = *thisptr++;  g1 = *thisptr++;  b1 = *thisptr++;
      RANGE(r1,0,255);  RANGE(g1,0,255);  RANGE(b1,0,255);  
      
      /* choose actual pixel value */
      val = (((r1&RMASK)>>RSHIFT) | ((g1&GMASK)>>GSHIFT) | 
       ((b1&BMASK)>>BSHIFT));
      *pp = val;
      
      /* compute color errors */
      r1 -= rmap[val];
      g1 -= gmap[val];
      b1 -= bmap[val];
      
      /* Add fractions of errors to adjacent pixels */
      if (j!=jmax) {  /* adjust RIGHT pixel */
  thisptr[0] += (r1*7) / 16;
  thisptr[1] += (g1*7) / 16;
  thisptr[2] += (b1*7) / 16;
      }
      
      if (i!=imax) {  /* do BOTTOM pixel */
  nextptr[0] += (r1*5) / 16;
  nextptr[1] += (g1*5) / 16;
  nextptr[2] += (b1*5) / 16;

  if (j>0) {  /* do BOTTOM LEFT pixel */
    nextptr[-3] += (r1*3) / 16;
    nextptr[-2] += (g1*3) / 16;
    nextptr[-1] += (b1*3) / 16;
  }

  if (j!=jmax) {  /* do BOTTOM RIGHT pixel */
    nextptr[3] += (r1)/16;
    nextptr[4] += (g1)/16;
    nextptr[5] += (b1)/16;
  }
  nextptr += 3;
      }
    }
  }
  
  free(thisline);
  free(nextline);
  return 0;


#undef RMASK
#undef RSHIFT
#undef GMASK
#undef GSHIFT
#undef BMASK
#undef BSHIFT
}
      






/***************************************************************/
/* The following code based on code from the 'pbmplus' package */
/* written by Jef Poskanzer                                    */
/***************************************************************/


/* ppmquant.c - quantize the colors in a pixmap down to a specified number
**
** Copyright (C) 1989, 1991 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

#define PPM_MAXMAXVAL 255

#define PPM_GETR(p) ((p).r)
#define PPM_GETG(p) ((p).g)
#define PPM_GETB(p) ((p).b)

#define PPM_ASSIGN(p,red,grn,blu) \
  { (p).r = (red); (p).g = (grn); (p).b = (blu); }

#define PPM_EQUAL(p,q) ( (p).r == (q).r && (p).g == (q).g && (p).b == (q).b )


/* Color scaling macro -- to make writing ppmtowhatever easier. */

#define PPM_DEPTH(newp,p,oldmaxval,newmaxval) \
    PPM_ASSIGN( (newp), \
          ((int) PPM_GETR(p)) * ((int)newmaxval) / ((int)oldmaxval), \
          ((int) PPM_GETG(p)) * ((int)newmaxval) / ((int)oldmaxval), \
          ((int) PPM_GETB(p)) * ((int)newmaxval) / ((int)oldmaxval) )


/* Luminance macro. */

/* 
 * #define PPM_LUMIN(p) \
 *   ( 0.299 * PPM_GETR(p) + 0.587 * PPM_GETG(p) + 0.114 * PPM_GETB(p) )
 */

/* Luminance macro, using only integer ops.  Returns an int (*256)  JHB */
#define PPM_LUMIN(p) \
  ( 77 * PPM_GETR(p) + 150 * PPM_GETG(p) + 29 * PPM_GETB(p) )




#define MAXCOLORS 32767
#define CLUSTER_MAXVAL 63

#define LARGE_LUM
#define REP_AVERAGE_PIXELS

#define FS_SCALE 1024

#define HASH_SIZE 6553

#define ppm_hashpixel(p) ((((int) PPM_GETR(p) * 33023 +    \
          (int) PPM_GETG(p) * 30013 +    \
          (int) PPM_GETB(p) * 27011) & 0x7fffffff)   \
        % HASH_SIZE)



/*** function defs ***/

static chist_vec   mediancut        (chist_vec, int, int, int, int);
static int         redcompare       (const void *, const void *);
static int         greencompare     (const void *, const void *);
static int         bluecompare      (const void *, const void *);
static int         sumcompare       (const void *, const void *);
static chash_table ppm_computechash (pixel **, int,int,int,int *);
static chash_table ppm_allocchash   (void);
static void        ppm_freechist    (chist_vec);
static void        ppm_freechash    (chash_table);


/****************************************************************************/
int IOfile::ppm_quant(const byte *pic24, int cols,int  rows, byte *pic8, 
                     byte *rmap, byte *gmap, byte *bmap, int newcolors)
{
  pixel**          pixels;
  register pixel*  pP;
  int              row;
  register int     col, limitcol;
  pixval           maxval, newmaxval;
  int              colors;
  register int     index;
  chist_vec chv, colormap;
  chash_table  cht;
  int              i;
  unsigned char    *picptr;
  static char      *fn = "ppmquant()";

  index = 0;
  maxval = 255;

  /*
   *  reformat 24-bit pic24 image (3 bytes per pixel) into 2-dimensional
   *  array of pixel structures
   */

  if (DEBUG) debug_printf("%s: remapping to ppm-style internal fmt\n", fn);
  WaitCursor();
  
  pixels = (pixel **) malloc(rows * sizeof(pixel *));
  if (!pixels) FatalError("couldn't allocate 'pixels' array");
  for (row=0; row<rows; row++) {
    pixels[row] = (pixel *) malloc(cols * sizeof(pixel));
    if (!pixels[row]) FatalError("couldn't allocate a row of pixels array");

    for (col=0, pP=pixels[row]; col<cols; col++, pP++) {
      pP->r = *pic24++;
      pP->g = *pic24++;
      pP->b = *pic24++;
    }
  }
  if (DEBUG) debug_printf("%s: done format remapping\n", fn);


    

  /*
   *  attempt to make a histogram of the colors, unclustered.
   *  If at first we don't succeed, lower maxval to increase color
   *  coherence and try again.  This will eventually terminate, with
   *  maxval at worst 15, since 32^3 is approximately MAXCOLORS.
   */

  WaitCursor();
  for ( ; ; ) {
    if (DEBUG) debug_printf( "%s:  making histogram\n", fn);

    chv = ppm_computechist(pixels, cols, rows, MAXCOLORS, &colors);
    if (chv != (chist_vec) 0) break;
    
    if (DEBUG) debug_printf( "%s: too many colors!\n", fn);
    newmaxval = maxval / 2;
    if (DEBUG) debug_printf( "%s: rescaling colors (maxval=%d) %s\n",
           fn, newmaxval, "to improve clustering");

    for (row=0; row<rows; ++row)
      for (col=0, pP=pixels[row]; col<cols; ++col, ++pP)
  PPM_DEPTH( *pP, *pP, maxval, newmaxval );
    maxval = newmaxval;
  }

  if (DEBUG) debug_printf("%s: %d colors found\n", fn, colors);



  /*
   * Step 3: apply median-cut to histogram, making the new colormap.
   */

  WaitCursor();
  if (DEBUG) debug_printf( "%s: choosing %d colors\n", fn, newcolors);
  colormap = mediancut(chv, colors, rows * cols, maxval, newcolors);
  ppm_freechist(chv);



  /*
   *  Step 4: map the colors in the image to their closest match in the
   *  new colormap, and write 'em out.
   */

  if (DEBUG) debug_printf("%s: mapping image to new colors\n", fn);
  cht = ppm_allocchash();

  picptr = pic8;
  for (row = 0;  row < rows;  ++row) {
    col = 0;  limitcol = cols;  pP = pixels[row];

    ProgressMeter(0, rows-1, row, "24 -> 8");

    if ((row & 0x1f) == 0) WaitCursor();
    do {
      int hash;
      chist_list chl;

      /* Check hash table to see if we have already matched this color. */

      hash = ppm_hashpixel(*pP);
      for (chl = cht[hash];  chl;  chl = chl->next)
  if (PPM_EQUAL(chl->ch.color, *pP)) {index = chl->ch.value; break;}

      if (!chl /*index = -1*/) {/* No; search colormap for closest match. */
  register int i, r1, g1, b1, r2, g2, b2;
  register long dist, newdist;

  r1 = PPM_GETR( *pP );
  g1 = PPM_GETG( *pP );
  b1 = PPM_GETB( *pP );
  dist = 2000000000;

  for (i=0; i<newcolors; i++) {
    r2 = PPM_GETR( colormap[i].color );
    g2 = PPM_GETG( colormap[i].color );
    b2 = PPM_GETB( colormap[i].color );

    newdist = ( r1 - r2 ) * ( r1 - r2 ) +
              ( g1 - g2 ) * ( g1 - g2 ) +
              ( b1 - b2 ) * ( b1 - b2 );

    if (newdist<dist) { index = i;  dist = newdist; }
  }

  hash = ppm_hashpixel(*pP);
  chl = (chist_list) malloc(sizeof(struct chist_list_item));
  if (!chl) FatalError("ran out of memory adding to hash table");

  chl->ch.color = *pP;
  chl->ch.value = index;
  chl->next = cht[hash];
  cht[hash] = chl;
      }

      *picptr++ = index;

      ++col;
      ++pP;
    }
    while (col != limitcol);
  }

  /* rescale the colormap and load the XV colormap */
  for (i=0; i<newcolors; i++) {
    PPM_DEPTH(colormap[i].color, colormap[i].color, maxval, 255);
    rmap[i] = PPM_GETR( colormap[i].color );
    gmap[i] = PPM_GETG( colormap[i].color );
    bmap[i] = PPM_GETB( colormap[i].color );
  }

  /* free the pixels array */
  for (i=0; i<rows; i++) free(pixels[i]);
  free(pixels);

  /* free cht and colormap */
  ppm_freechist(colormap);
  ppm_freechash(cht);

  return 0;
}



/*
** Here is the fun part, the median-cut colormap generator.  This is based
** on Paul Heckbert's paper "Color Image Quantization for Frame Buffer
** Display", SIGGRAPH '82 Proceedings, page 297.
*/



/****************************************************************************/
chist_vec IOfile::mediancut( chist_vec chv, int colors, int sum, int maxval, int newcolors )
{
  chist_vec colormap;
  box_vector bv;
  register int bi, i;
  int boxes;

  bv = (box_vector) malloc(sizeof(struct box) * newcolors);
  colormap = (chist_vec) 
             malloc(sizeof(struct chist_item) * newcolors );

  if (!bv || !colormap) FatalError("unable to malloc in mediancut()");

  for (i=0; i<newcolors; i++)
    PPM_ASSIGN(colormap[i].color, 0, 0, 0);

  /*
   *  Set up the initial box.
   */
  bv[0].index = 0;
  bv[0].colors = colors;
  bv[0].sum = sum;
  boxes = 1;


  /*
   ** Main loop: split boxes until we have enough.
   */

  while ( boxes < newcolors ) {
    register int indx, clrs;
    int sm;
    register int minr, maxr, ming, maxg, minb, maxb, v;
    int halfsum, lowersum;

    /*
     ** Find the first splittable box.
     */
    for (bi=0; bv[bi].colors<2 && bi<boxes; bi++) ;
    if (bi == boxes) break;  /* ran out of colors! */

    indx = bv[bi].index;
    clrs = bv[bi].colors;
    sm = bv[bi].sum;

    /*
     ** Go through the box finding the minimum and maximum of each
     ** component - the boundaries of the box.
     */
    minr = maxr = PPM_GETR( chv[indx].color );
    ming = maxg = PPM_GETG( chv[indx].color );
    minb = maxb = PPM_GETB( chv[indx].color );

    for (i=1; i<clrs; i++) {
      v = PPM_GETR( chv[indx + i].color );
      if (v < minr) minr = v;
      if (v > maxr) maxr = v;

      v = PPM_GETG( chv[indx + i].color );
      if (v < ming) ming = v;
      if (v > maxg) maxg = v;

      v = PPM_GETB( chv[indx + i].color );
      if (v < minb) minb = v;
      if (v > maxb) maxb = v;
    }

    /*
     ** Find the largest dimension, and sort by that component.  I have
     ** included two methods for determining the "largest" dimension;
     ** first by simply comparing the range in RGB space, and second
     ** by transforming into luminosities before the comparison.  You
     ** can switch which method is used by switching the commenting on
     ** the LARGE_ defines at the beginning of this source file.
     */
    {
      /* LARGE_LUM version */

      pixel p;
      int rl, gl, bl;

      PPM_ASSIGN(p, maxr - minr, 0, 0);
      rl = PPM_LUMIN(p);

      PPM_ASSIGN(p, 0, maxg - ming, 0);
      gl = PPM_LUMIN(p);

      PPM_ASSIGN(p, 0, 0, maxb - minb);
      bl = PPM_LUMIN(p);

      if (rl >= gl && rl >= bl)
  qsort((char*) &(chv[indx]), (size_t) clrs, sizeof(struct chist_item),
        redcompare );
      else if (gl >= bl)
  qsort((char*) &(chv[indx]), (size_t) clrs, sizeof(struct chist_item),
        greencompare );
      else 
  qsort((char*) &(chv[indx]), (size_t) clrs, sizeof(struct chist_item),
        bluecompare );
    }

    /*
     ** Now find the median based on the counts, so that about half the
     ** pixels (not colors, pixels) are in each subdivision.
     */
    lowersum = chv[indx].value;
    halfsum = sm / 2;
    for (i=1; i<clrs-1; i++) {
      if (lowersum >= halfsum) break;
      lowersum += chv[indx + i].value;
    }

    /*
     ** Split the box, and sort to bring the biggest boxes to the top.
     */
    bv[bi].colors = i;
    bv[bi].sum = lowersum;
    bv[boxes].index = indx + i;
    bv[boxes].colors = clrs - i;
    bv[boxes].sum = sm - lowersum;
    ++boxes;
    qsort((char*) bv, (size_t) boxes, sizeof(struct box), sumcompare);
  }  /* while (boxes ... */
  
  /*
   ** Ok, we've got enough boxes.  Now choose a representative color for
   ** each box.  There are a number of possible ways to make this choice.
   ** One would be to choose the center of the box; this ignores any structure
   ** within the boxes.  Another method would be to average all the colors in
   ** the box - this is the method specified in Heckbert's paper.  A third
   ** method is to average all the pixels in the box.  You can switch which
   ** method is used by switching the commenting on the REP_ defines at
   ** the beginning of this source file.
   */
  
  for (bi=0; bi<boxes; bi++) {
    /* REP_AVERAGE_PIXELS version */
    register int indx = bv[bi].index;
    register int clrs = bv[bi].colors;
    register long r = 0, g = 0, b = 0, sum = 0;

    for (i=0; i<clrs; i++) {
      r += PPM_GETR( chv[indx + i].color ) * chv[indx + i].value;
      g += PPM_GETG( chv[indx + i].color ) * chv[indx + i].value;
      b += PPM_GETB( chv[indx + i].color ) * chv[indx + i].value;
      sum += chv[indx + i].value;
    }

    r = r / sum;  if (r>maxval) r = maxval;  /* avoid math errors */
    g = g / sum;  if (g>maxval) g = maxval;
    b = b / sum;  if (b>maxval) b = maxval;

    PPM_ASSIGN( colormap[bi].color, (byte) r, (byte) g, (byte) b );
  }

  free(bv);
  return colormap;
}


/**********************************/
static int redcompare(const void *p1, const void *p2)
{
  return (int) PPM_GETR( ((chist_vec)p1)->color ) - 
         (int) PPM_GETR( ((chist_vec)p2)->color );
}

/**********************************/
static int greencompare(const void *p1, const void *p2)
{
  return (int) PPM_GETG( ((chist_vec)p1)->color ) - 
         (int) PPM_GETG( ((chist_vec)p2)->color );
}

/**********************************/
static int bluecompare(const void *p1, const void *p2)
{
  return (int) PPM_GETB( ((chist_vec)p1)->color ) - 
         (int) PPM_GETB( ((chist_vec)p2)->color );
}

/**********************************/
static int sumcompare( const void *p1, const void *p2)
{
  return ((box_vector) p2)->sum - ((box_vector) p1)->sum;
}



/****************************************************************************/
chist_vec IOfile::ppm_computechist(pixel** pixels, int cols, int rows,
                                      int maxcolors, int* colorsP)
{
  chash_table cht;
  chist_vec chv;

  cht = ppm_computechash(pixels, cols, rows, maxcolors, colorsP);
  if (!cht) return (chist_vec) 0;

  chv = ppm_chashtochist(cht, maxcolors);
  ppm_freechash(cht);
  return chv;
}


/****************************************************************************/
chash_table IOfile::ppm_computechash(pixel** pixels, int cols, int rows, 
              int maxcolors,int*  colorsP )
{
  chash_table cht;
  register pixel* pP;
  chist_list chl;
  int col, row, hash;

  cht = ppm_allocchash( );
  *colorsP = 0;

  /* Go through the entire image, building a hash table of colors. */
  for (row=0; row<rows; row++)
    for (col=0, pP=pixels[row];  col<cols;  col++, pP++) {
      hash = ppm_hashpixel(*pP);

      for (chl = cht[hash]; chl != (chist_list) 0; chl = chl->next)
  if (PPM_EQUAL(chl->ch.color, *pP)) break;
      
      if (chl != (chist_list) 0) ++(chl->ch.value);
      else {
  if ((*colorsP)++ > maxcolors) {
    ppm_freechash(cht);
    return (chash_table) 0;
  }
  
  chl = (chist_list) malloc(sizeof(struct chist_list_item));
  if (!chl) FatalError("ran out of memory computing hash table");

  chl->ch.color = *pP;
  chl->ch.value = 1;
  chl->next = cht[hash];
  cht[hash] = chl;
      }
    }
  
  return cht;
}


/****************************************************************************/
chash_table IOfile::ppm_allocchash()
{
  chash_table cht;
  int i;

  cht = (chash_table) malloc( HASH_SIZE * sizeof(chist_list) );
  if (!cht) FatalError("ran out of memory allocating hash table");

  for (i=0; i<HASH_SIZE; i++ )
    cht[i] = (chist_list) 0;

  return cht;
}


/****************************************************************************/
chist_vec IOfile::ppm_chashtochist( chash_table cht, int maxcolors )
{
  chist_vec chv;
  chist_list chl;
  int i, j;

  /* Now collate the hash table into a simple chist array. */
  chv = (chist_vec) malloc( maxcolors * sizeof(struct chist_item) );

  /* (Leave room for expansion by caller.) */
  // We dont use object calls since we dont have this parameter.
  if (!chv) FatalError("ran out of memory generating histogram");

  /* Loop through the hash table. */
  j = 0;
  for (i=0; i<HASH_SIZE; i++)
    for (chl = cht[i];  chl != (chist_list) 0;  chl = chl->next) {
      /* Add the new entry. */
      chv[j] = chl->ch;
      ++j;
    }

  return chv;
}


/****************************************************************************/
static void ppm_freechist(chist_vec chv )
{
  free( (char*) chv );
}


/****************************************************************************/
static void ppm_freechash(chash_table  cht )
{
  int i;
  chist_list chl, chlnext;

  for (i=0; i<HASH_SIZE; i++)
    for (chl = cht[i];  chl != (chist_list) 0; chl = chlnext) {
      chlnext = chl->next;
      free( (char*) chl );
    }

  free( (char*) cht );
}


/*************************************************************/
/* ALGORITHM TO GENERATE ONLY THE COLOR MAP OF A RGB PICTURE */
/*************************************************************/
int quick_check_map(const byte *pic24, int w,int h, byte *rmap,
                       byte *gmap,byte *bmap, int maxcol)
{
  /* scans picture until it finds more than 'maxcol' different colors.  If it
     finds more than 'maxcol' colors, it returns '0'.  If it DOESN'T, it does
     the 24-to-8 conversion by simply sticking the colors it found into
     a colormap, and changing instances of a color in pic24 into colormap
     indicies (in pic8) */

  unsigned long colors[256],col;
  int           i, nc, low, high, mid;
  byte         *p;

  if (maxcol>256) maxcol = 256;

  /* put the first color in the table by hand */
  nc = 0;  mid = 0;  

  for (i=w*h,p=(byte*)pic24; i; i--) {
    col  = (((u_long) *p++) << 16);  
    col += (((u_long) *p++) << 8);
    col +=  *p++;

    /* binary search the 'colors' array to see if it's in there */
    low = 0;  high = nc-1;
    while (low <= high) {
      mid = (low+high)/2;
      if      (col < colors[mid]) high = mid - 1;
      else if (col > colors[mid]) low  = mid + 1;
      else break;
    }

    if (high < low) { /* didn't find color in list, add it. */
      if (nc>=maxcol) return 0;
      xvbcopy((char *) &colors[low], (char *) &colors[low+1],
        (nc - low) * sizeof(u_long));
      colors[low] = col;
      nc++;
    }
  }


  /* and load up the 'desired colormap' */
  for (i=0; i<nc; i++) {
    rmap[i] = (byte) (colors[i]>>16);  
    gmap[i] = (byte) (colors[i]>>8) & 0xff;
    bmap[i] = (byte) colors[i]      & 0xff;
  }

  return 1;
}


/****************************/
int GetSlowMapFromPIC24(const byte *pic24,int w,int h,int nc,
                           byte *rm,byte *gm,byte *bm)
{
  /* returns 0 if successfull, -1 if error.
   * this function computes colormap of nc number of color in 
   * rm (red map) , gm (green map) and bm (blue map)
   * uses slow method */
  
  int   i;
  
  if (!pic24) return -1;

  if (nc<=0) nc = 256;  /* 'nc == 0' breaks code */

  if (quick_check_map(pic24, w,h, rm,gm,bm, nc)) { 
    return 0;   
  }

  i = slow_quant_map(pic24, w, h, rm, gm, bm, nc);
  
  if (i) { 
   return -1;
  }
  return 0;
}


